##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################
__version__='$Revision: 1.5 $'[11:-2]

from string import join, split, find, rfind, lower, upper
from urllib import quote

_marker=[]
class BaseRequest:
    """Provide basic ZPublisher request management
    
    This object provides access to request data. Request data may
    vary depending on the protocol used.

    Request objects are created by the object publisher and will be
    passed to published objects through the argument name, REQUEST.

    The request object is a mapping object that represents a
    collection of variable to value mappings.
    """

    _file=None
    _auth=None
    
    def __init__(self, other=None, **kw):
        """The constructor is not allowed to raise errors
        """
        if other is None: other=kw
        else: other.update(kw)
        self.other=other

    def close(self):
        self.other.clear()

    def processInputs(self):
        """Do any input processing that could raise errors
        """

    def __len__(self):
        return 1

    def __setitem__(self,key,value):
        """Set application variables

        This method is used to set a variable in the requests "other"
        category.
        """
        self.other[key]=value

    set=__setitem__

    def __getitem__(self,key,
                    default=_marker, # Any special internal marker will do
                    ):
        """Get a variable value

        Return a value for the required variable name.
        The value will be looked up from one of the request data
        categories. The search order is environment variables,
        other variables, form data, and then cookies. 
        
        """
        v=self.other.get(key, _marker)
        if v is not _marker: return v

        if key=='BODY' and self._file is not None:
            p=self._file.tell()
            self._file.seek(0)
            v=self._file.read()
            self._file.seek(p)
            self.other[key]=v
            return v

        if key=='BODYFILE' and self._file is not None:
            v=self._file
            self.other[key]=v
            return v
        
        raise KeyError, key

    __getattr__=get=__getitem__

    def has_key(self,key):
        return self.get(key, _marker) is not _marker

    def keys(self):
        keys = {}
        keys.update(self.other)
        return keys.keys()

    def items(self):
        result = []
        get=self.get
        for k in self.keys():
            result.append((k, get(k)))
        return result

    def values(self):
        result = []
        get=self.get
        for k in self.keys():
            result.append(get(k))
        return result



"""quixote.errors

Exception classes used by Quixote
"""

class IgnoredException (Exception):
    """Handlers can raise this exception to bail out early; Quixote will
    ignore the exception and simply perform its usual processing.
    """
    
class AccessError (Exception):
    """Raised when a user's access to a given URL is denied.
    """

    def __init__ (self, msg=None):
        self.msg = msg or "access denied"
    
    def __str__ (self):
        return self.msg

    
class TraversalError (Exception):
    """Raised when a URL attempts to traverse to a bogus or non-
    existent Python namespace.  There are a number of reasons why this
    can happen, so these exceptions should have a clear and explicit
    message explaining the problem.  However, that can reveal valuable
    information to attackers, so the '__str__()' method here will hide
    that useful information if the configuration variable SECURE_ERRORS
    is true.
    """

    # XXX there should be some interaction between the response object
    # and this exception, eg. the ability to generate a pretty HTML
    # document (with or without traceback) and an appropriate HTTP
    # response code (most likely 404).

    # XXX hiding the exact error message is kind of pointless if
    # we show tracebacks, since a line of code like
    #   raise TraversalError("here's what's really going on", url)
    # might appear at the end of the traceback.  Thus, "secure errors"
    # aren't really secure unless tracebacks are also disabled.

    def __init__ (self, msg, url):
        self.msg = msg
        self.url = url

    def __str__ (self):
        from quixote import config
        if config.SECURE_ERRORS:
            return "not found: %s" % self.url
        else:
            return self.msg



##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################

__version__='$Revision: 1.11 $'[11:-2]

import regex, re, sys, os, string
from string import lower, atoi, rfind, split, strip, join, upper, find
from BaseRequest import BaseRequest
from HTTPResponse import HTTPResponse
from cgi import FieldStorage
from urllib import quote, unquote
xmlrpc=None # Placeholder for module that we'll import if we have to.

isCGI_NAME = {
        'SERVER_SOFTWARE' : 1, 
        'SERVER_NAME' : 1, 
        'GATEWAY_INTERFACE' : 1, 
        'SERVER_PROTOCOL' : 1, 
        'SERVER_PORT' : 1, 
        'REQUEST_METHOD' : 1, 
        'PATH_INFO' : 1, 
        'PATH_TRANSLATED' : 1, 
        'SCRIPT_NAME' : 1, 
        'QUERY_STRING' : 1, 
        'REMOTE_HOST' : 1, 
        'REMOTE_ADDR' : 1, 
        'AUTH_TYPE' : 1, 
        'REMOTE_USER' : 1, 
        'REMOTE_IDENT' : 1, 
        'CONTENT_TYPE' : 1, 
        'CONTENT_LENGTH' : 1,
        'SERVER_URL': 1,
        }.has_key

hide_key={'HTTP_AUTHORIZATION':1,
          'HTTP_CGI_AUTHORIZATION': 1,
          }.has_key

_marker=[]
class HTTPRequest(BaseRequest):
    """\
    Model HTTP request data.
    
    This object provides access to request data.  This includes, the
    input headers, form data, server data, and cookies.

    Request objects are created by the object publisher and will be
    passed to published objects through the argument name, REQUEST.

    The request object is a mapping object that represents a
    collection of variable to value mappings.  In addition, variables
    are divided into four categories:

      - Environment variables

        These variables include input headers, server data, and other
        request-related data.  The variable names are as <a
        href="http://hoohoo.ncsa.uiuc.edu/cgi/env.html">specified</a>
        in the <a
        href="http://hoohoo.ncsa.uiuc.edu/cgi/interface.html">CGI
        specification</a>

      - Form data

        These are data extracted from either a URL-encoded query
        string or body, if present.

      - Cookies

        These are the cookie data, if present.

      - Other

        Data that may be set by an application object.

    The form attribute of a request is actually a Field Storage
    object.  When file uploads are used, this provides a richer and
    more complex interface than is provided by accessing form data as
    items of the request.  See the FieldStorage class documentation
    for more details.

    The request object may be used as a mapping object, in which case
    values will be looked up in the order: environment variables,
    other variables, form data, and then cookies.
    """
    _hacked_path=None
    args=()
    _file=None
    _urls = ()

    retry_max_count=3
    def supports_retry(self): return self.retry_count < self.retry_max_count

    def retry(self):
        self.retry_count=self.retry_count+1
        self.stdin.seek(0)
        r=self.__class__(stdin=self.stdin,
                         environ=self._orig_env,
                         response=self.response.retry()
                         )
        return r


    def setSite(self, base=None):
        """ blah """
        if base is not None:
            self.script = self.base = base

        # reset the URL
        self.other['URL'] = self.script
        # clear the 'cache' of URLx and BASEx
        for x in self._urls:
            del self.other[x]
            
        del self.other['PARENTS'][:]


    def __init__(self, stdin, environ, response):
        self._orig_env=environ

        if environ.has_key('HTTP_AUTHORIZATION'):
            self._auth=environ['HTTP_AUTHORIZATION']
            response._auth=1
            del environ['HTTP_AUTHORIZATION']

        self.stdin=stdin
        self.environ=environ
        have_env=environ.has_key
        get_env=environ.get
        self.response=response
        other=self.other={'RESPONSE': response}
        self.form={}
        self.steps=[]

        ################################################################
        # Get base info first. This isn't likely to cause
        # errors and might be useful to error handlers.
        b=script=strip(get_env('SCRIPT_NAME',''))
        while b and b[-1]=='/': b=b[:-1]
        p = rfind(b,'/')
        if p >= 0: b=b[:p+1]
        else: b=''
        while b and b[0]=='/': b=b[1:]

        server_url=get_env('SERVER_URL',None)
        if server_url is not None:
             server_url=strip(server_url)
        else:
             if have_env('HTTPS') and (
                 environ['HTTPS'] == "on" or environ['HTTPS'] == "ON"):
                 server_url='https://'
             elif (have_env('SERVER_PORT_SECURE') and 
                   environ['SERVER_PORT_SECURE'] == "1"):
                 server_url='https://'
             else: server_url='http://'

             if have_env('HTTP_HOST'):
                 server_url=server_url+strip(environ['HTTP_HOST'])
             else:
                 server_url=server_url+strip(environ['SERVER_NAME'])
                 server_port=environ['SERVER_PORT']
                 if server_port!='80': server_url=server_url+':'+server_port
             other['SERVER_URL']=server_url
             
        if server_url[-1:]=='/': server_url=server_url[:-1]
                        
        if b: self.base="%s/%s" % (server_url,b)
        else: self.base=server_url
        while script[:1]=='/': script=script[1:]
        if script: script="%s/%s" % (server_url,script)
        else:      script=server_url
        other['URL']=self.script=script

        ################################################################
        # Cookie values should *not* be appended to existing form
        # vars with the same name - they are more like default values
        # for names not otherwise specified in the form.
        cookies={}
        k=get_env('HTTP_COOKIE','')
        if k:
            parse_cookie(k, cookies)
            for k,item in cookies.items():
                if not other.has_key(k):
                    other[k]=item
        self.cookies=cookies

        self.session = None
        
    def processInputs(
        self,
        # "static" variables that we want to be local for speed
        SEQUENCE=1,
        DEFAULT=2,
        RECORD=4,
        RECORDS=8,
        REC=12, # RECORD|RECORDS
        EMPTY=16,
        CONVERTED=32,
        hasattr=hasattr,
        getattr=getattr,
        setattr=setattr,
        search_type=regex.compile(':[a-zA-Z][a-zA-Z0-9_]+$').search,
        rfind=string.rfind,
        ):
        """Process request inputs

        We need to delay input parsing so that it is done under
        publisher control for error handling purposes.
        """
        response=self.response
        environ=self.environ
        method=environ.get('REQUEST_METHOD','GET')
        
        if method != 'GET': fp=self.stdin
        else:               fp=None

        form=self.form
        other=self.other

        meth=None
        fs=FieldStorage(fp=fp,environ=environ,keep_blank_values=1)
        if not hasattr(fs,'list') or fs.list is None:
            # Hm, maybe it's an XML-RPC
            if (fs.headers.has_key('content-type') and
                fs.headers['content-type'] == 'text/xml' and
                method == 'POST'):
                # Ye haaa, XML-RPC!
                global xmlrpc
                if xmlrpc is None: import xmlrpc
                meth, self.args = xmlrpc.parse_input(fs.value)
                response=xmlrpc.response(response)
                other['RESPONSE']=self.response=response
                other['REQUEST_METHOD']='' # We don't want index_html!
            else:
                self._file=fs.file
        else:
            fslist=fs.list
            tuple_items={}
            lt=type([])
            CGI_name=isCGI_NAME
            defaults={}
            converter=seqf=None
            
            for item in fslist:
                
                key=item.name
                if (hasattr(item,'file') and hasattr(item,'filename')
                    and hasattr(item,'headers')):
                    if (item.file and
                        (item.filename is not None
                         # RFC 1867 says that all fields get a content-type.
                         # or 'content-type' in map(lower, item.headers.keys())
                         )):
                        item=FileUpload(item)
                    else:
                        item=item.value

                flags=0

                # Loop through the different types and set
                # the appropriate flags

                # We'll search from the back to the front.
                # We'll do the search in two steps.  First, we'll
                # do a string search, and then we'll check it with
                # a regex search.
                
                l=rfind(key,':')
                if l >= 0:
                    l=search_type(key,l)
                    while l >= 0:
                        type_name=key[l+1:]
                        key=key[:l]
                        if type_name == 'list':
                            seqf=list
                            flags=flags|SEQUENCE
                        elif type_name == 'tuple':
                            seqf=tuple
                            tuple_items[key]=1
                            flags=flags|SEQUENCE
                        elif type_name == 'method':
                            if l: meth=key
                            else: meth=item
                        elif type_name == 'default_method':
                            if not meth:
                                if l: meth=key
                                else: meth=item
                        elif type_name == 'default':
                            flags=flags|DEFAULT
                        elif type_name == 'record':
                            flags=flags|RECORD
                        elif type_name == 'records':
                            flags=flags|RECORDS
                        elif type_name == 'ignore_empty':
                            if not item: flags=flags|EMPTY
    
                        l=rfind(key,':')
                        if l < 0: break
                        l=search_type(key,l)
             
                # Filter out special names from form:
                if CGI_name(key) or key[:5]=='HTTP_': continue

                if flags:

                    # skip over empty fields    
                    if flags&EMPTY: continue

                    #Split the key and its attribute
                    if flags&REC:
                        key=split(key,".")
                        key, attr=join(key[:-1],"."), key[-1]
                       
                    # defer conversion
                    if flags&CONVERTED:
                        try:
                            item=converter(item)
                        except:
                            if (not item and not (flags&DEFAULT) and
                                defaults.has_key(key)):
                                item = defaults[key]
                                if flags&RECORD:
                                    item=getattr(item,attr)
                                if flags&RECORDS:
                                    item.reverse()
                                    item = item[0]
                                    item=getattr(item,attr)
                            else:
                                raise                            
                         
                    #Determine which dictionary to use
                    if flags&DEFAULT:
                       mapping_object = defaults
                    else:
                       mapping_object = form

                    #Insert in dictionary
                    if mapping_object.has_key(key):
                       if flags&RECORDS:
                           #Get the list and the last record
                           #in the list
                           reclist = mapping_object[key]
                           reclist.reverse()
                           x=reclist[0]
                           reclist.reverse()
                           if not hasattr(x,attr):
                               #If the attribute does not
                               #exist, set it
                               if flags&SEQUENCE: item=[item]
                               reclist.remove(x)
                               setattr(x,attr,item)
                               reclist.append(x)
                               mapping_object[key] = reclist
                           else:
                               if flags&SEQUENCE:
                                   # If the attribute is a
                                   # sequence, append the item
                                   # to the existing attribute
                                   reclist.remove(x)
                                   y = getattr(x, attr)
                                   y.append(item)
                                   setattr(x, attr, y)
                                   reclist.append(x)
                                   mapping_object[key] = reclist
                               else:
                                   # Create a new record and add
                                   # it to the list
                                   n=record()
                                   setattr(n,attr,item)
                                   reclist.append(n)
                                   mapping_object[key]=reclist
                       elif flags&RECORD:
                           b=mapping_object[key]
                           if flags&SEQUENCE:
                              item=[item]
                              if not hasattr(b,attr):
                                  # if it does not have the
                                  # attribute, set it
                                  setattr(b,attr,item)
                              else:
                                  # it has the attribute so
                                  # append the item to it
                                  setattr(b,attr,getattr(b,attr)+item)
                           else:
                              # it is not a sequence so
                              # set the attribute
                              setattr(b,attr,item)        
                       else:
                          # it is not a record or list of records
                           found=mapping_object[key]
                           if type(found) is lt:
                               found.append(item)
                           else:
                               found=[found,item]
                               mapping_object[key]=found
                    else:
                       # The dictionary does not have the key
                       if flags&RECORDS:
                           # Create a new record, set its attribute
                           # and put it in the dictionary as a list
                           a = record()
                           if flags&SEQUENCE: item=[item]
                           setattr(a,attr,item)
                           mapping_object[key]=[a]
                       elif flags&RECORD:
                           # Create a new record, set its attribute
                           # and put it in the dictionary
                           if flags&SEQUENCE: item=[item]
                           r = mapping_object[key]=record()
                           setattr(r,attr,item)
                       else:
                           # it is not a record or list of records
                           if flags&SEQUENCE: item=[item]
                           mapping_object[key]=item

                else:
                    # This branch is for case when no type was specified.
                    mapping_object = form
    
                    #Insert in dictionary
                    if mapping_object.has_key(key):
                        # it is not a record or list of records
                        found=mapping_object[key]
                        if type(found) is lt:
                            found.append(item)
                        else:
                            found=[found,item]
                            mapping_object[key]=found
                    else:
                        mapping_object[key]=item

            #insert defaults into form dictionary
            if defaults:
                for keys, values in defaults.items():
                    if not form.has_key(keys):
                        # if the form does not have the key,
                        # set the default
                        form[keys]=values
                    else:
                        #The form has the key
                        if getattr(values, '__class__',0) is record:
                           # if the key is mapped to a record, get the
                           # record
                           r = form[keys]
                           for k, v in values.__dict__.items():
                              # loop through the attributes and values
                              # in the default dictionary
                              if not hasattr(r, k):
                                 # if the form dictionary doesn't have
                                 # the attribute, set it to the default
                                 setattr(r,k,v)
                                 form[keys] = r    
                        elif values == type([]):
                               # the key is mapped to a list
                               l = form[keys]
                               for x in values:
                                   # for each x in the list
                                   if getattr(x, '__class__',0) is record:
                                       # if the x is a record
                                       for k, v in x.__dict__.items():
                                           
                                           # loop through each
                                           # attribute and value in
                                           # the record
                                           
                                           for y in l:
                                               
                                               # loop through each
                                               # record in the form
                                               # list if it doesn't
                                               # have the attributes
                                               # in the default
                                               # dictionary, set them
                                               
                                               if not hasattr(y, k):
                                                   setattr(y, k, v)
                                   else:
                                       # x is not a record
                                       if not a in l:
                                           l.append(a)
                               form[keys] = l
                        else:
                            # The form has the key, the key is not mapped
                            # to a record or sequence so do nothing
                            pass
                                
            # Convert to tuples
            if tuple_items:
                for key in tuple_items.keys():
                   # Split the key and get the attr
                   k=split(key, ".")
                   k,attr=join(k[:-1], "."), k[-1]
                   a = attr
                   # remove any type_names in the attr
                   while not a=='':
                      a=split(a, ":")
                      a,new=join(a[:-1], ":"), a[-1]
                   attr = new
                   if form.has_key(k):
                      # If the form has the split key get its value
                      item =form[k]
                      if (hasattr(item, '__class__') and
                          item.__class__ is record):
                         # if the value is mapped to a record, check if it
                         # has the attribute, if it has it, convert it to
                         # a tuple and set it
                         if hasattr(item,attr):
                            value=tuple(getattr(item,attr))
                            setattr(item,attr,value)
                      else:
                         # It is mapped to a list of  records
                         for x in item:
                            # loop through the records
                            if hasattr(x, attr):
                               # If the record has the attribute
                               # convert it to a tuple and set it
                               value=tuple(getattr(x,attr))
                               setattr(x,attr,value)          
                   else:
                      # the form does not have the split key 
                      if form.has_key(key):
                         # if it has the original key, get the item
                         # convert it to a tuple
                         item=form[key]  
                         item=tuple(form[key])
                         form[key]=item
                     
        other.update(form)
        if meth:
            if environ.has_key('PATH_INFO'):
                path=environ['PATH_INFO']
                while path[-1:]=='/': path=path[:-1]
            else: path=''
            other['PATH_INFO']=path="%s/%s" % (path,meth)
            self._hacked_path=1

    def get_header(self, name, default=None):
        """Return the named HTTP header, or an optional default
        argument or None if the header is not found. Note that
        both original and CGI-ified header names are recognized,
        e.g. 'Content-Type', 'CONTENT_TYPE' and 'HTTP_CONTENT_TYPE'
        should all return the Content-Type header, if available.
        """
        environ=self.environ
        name=upper(join(split(name,"-"),"_"))
        val=environ.get(name, None)
        if val is not None:
            return val
        if name[:5] != 'HTTP_':
            name='HTTP_%s' % name
        return environ.get(name, default)


    def __getitem__(self,key,
                    default=_marker, # Any special internal marker will do
                    URLmatch=regex.compile('URL[0-9]+$').match,
                    BASEmatch=regex.compile('BASE[0-9]+$').match,
                    ):
        """Get a variable value

        Return a value for the required variable name.
        The value will be looked up from one of the request data
        categories. The search order is environment variables,
        other variables, form data, and then cookies. 
        
        """ #"
        other=self.other
        if other.has_key(key):
            return other[key]

        if key[:1]=='U' and URLmatch(key) >= 0:
            n=atoi(key[3:])
            URL=other['URL']
            for i in range(0,n):
                l=rfind(URL,'/')
                if l >= 0: URL=URL[:l]
                else: raise KeyError, key
                if len(URL) < len(self.base) and n > 1: raise KeyError, key
            other[key]=URL
            self._urls = self._urls + (key,)
            return URL

        if isCGI_NAME(key) or key[:5] == 'HTTP_':
            environ=self.environ
            if environ.has_key(key) and (not hide_key(key)):
                return environ[key]
            return ''

        if key[:1]=='B':
            if BASEmatch(key) >= 0:
                n=atoi(key[4:])
                if n:
                    n=n-1
                    
                    if len(self.steps) < n:
                        raise KeyError, key

                    v=self.script
                    while v[-1:]=='/': v=v[:-1]
                    v=join([v]+self.steps[:n],'/')
                else:
                    v=self.base
                    while v[-1:]=='/': v=v[:-1]
                other[key]=v
                self._urls = self._urls + (key,)
                return v

            if key=='BODY' and self._file is not None:
                p=self._file.tell()
                self._file.seek(0)
                v=self._file.read()
                self._file.seek(p)
                self.other[key]=v
                return v

            if key=='BODYFILE' and self._file is not None:
                v=self._file
                self.other[key]=v
                return v

        if default is not _marker: return default

        raise KeyError, key

    __getattr__=__getitem__

    def get(self, key, default=None):
        return self.__getitem__(key, default)

    def has_key(self, key):
        try: self[key]
        except: return 0
        else: return 1

    def keys(self):
        keys = {}

        for key in self.environ.keys():
            if (isCGI_NAME(key) or key[:5] == 'HTTP_') and \
               (not hide_key(key)):
                    keys[key] = 1

        n=0
        while 1:
            n=n+1
            key = "URL%s" % n
            if not self.has_key(key): break

        n=0
        while 1:
            n=n+1
            key = "BASE%s" % n
            if not self.has_key(key): break

        keys.update(self.other)

        keys=keys.keys()
        keys.sort()

        return keys

    def dump_html(self):
        result="<h3>form</h3><table>"
        row='<tr valign="top" align="left"><th>%s</th><td>%s</td></tr>'
        for k,v in self.form.items():
            result=result + row % (html_quote(k), html_quote(v))
        result=result+"</table><h3>cookies</h3><table>"
        for k,v in self.cookies.items():
            result=result + row % (html_quote(k), html_quote(v))
        result=result+"</table><h3>other</h3><table>"
        for k,v in self.other.items():
            if k in ('PARENTS','RESPONSE'): continue
            result=result + row % (html_quote(k), html_quote(v))
    
        for n in "0123456789":
            key = "URL%s"%n
            try: result=result + row % (key, html_quote(self[key])) 
            except KeyError: pass
        for n in "0123456789":
            key = "BASE%s"%n
            try: result=result + row % (key, html_quote(self[key])) 
            except KeyError: pass

        result=result+"</table><h3>environ</h3><table>"
        for k,v in self.environ.items():
            if not hide_key(k):
                result=result + row % (html_quote(k), html_quote(v))
        return result+"</table>"

    def dump(self):
        result=""
        row='%-15s %s\n'
        L = self.form.items() ; L.sort()
        for k,v in L:
            result=result + row % (k,v)
        result=result+"\nCookies:\n"
        L = self.cookies.items() ; L.sort()
        for k,v in L:
            result=result + row % (k,v)
        result=result+"\nOther:\n"
        L = self.other.items() ; L.sort()
        for k,v in L:
            if k in ('PARENTS','RESPONSE'): continue
            result=result + row % (k,v)

        for n in "0123456789":
            key = "URL%s"%n
            try: result=result + row % (key,self[key])
            except KeyError: pass
        for n in "0123456789":
            key = "BASE%s"%n
            try: result=result + row % (key,self[key])
            except KeyError: pass

        result=result+"\nEnvironment:\n"
        L = self.environ.items() ; L.sort()
        for k,v in L:
            if not hide_key(k):
                result=result + row % (k,v)
        return result

    def browser_version (self):
        """Examine the User-agent request header to try to figure out
           what the current browser is.  Returns a (name,version) tuple
           where each element is a string or None if we couldn't parse the
           User-agent header.  Note that Netscape {Communicator,Navigator}
           and Internet Explorer advertise themselves as, respectively,
           "Mozilla" and "MSIE"."""

        ua = self.get_header ('user-agent')
        if ua is None:
            return (None, None)

        # Looks like Netscape...
        moz_match = re.search (r'(Mozilla)/([\w\.]+)', ua)
        if moz_match:
            # ...but smells like Microsoft
            ie_match = re.search (r'(MSIE)\s+([\w\.]+)', ua)
            if ie_match:
                return ie_match.group (1, 2)

            # nope, it's really Netscape
            else:
                return moz_match.group (1, 2)

        # neither Netscape or IE -- try a broader regex
        else:
            match = re.search (r'([^\s\/]+)/([\w\.]+)', ua)
            if match:
                return match.group (1, 2)
            else:
                return (None, None)

    # browser_version ()
    
    def _authUserPW(self):
        global base64
        auth=self._auth
        if auth:
            if lower(auth[:6]) == 'basic ':
                if base64 is None: import base64
                [name,password] = split(
                    base64.decodestring(split(auth)[-1]), ':')
                return name, password




base64=None


# This is duplicated from DocumentTemplate.DT_Util to
# prevent a dependency on the DocumentTemplate package.
# Some folks still use the ZPublisher package as a
# standalone publisher without DocumentTemplate.
def html_quote(value, character_entities=(
                      (('&'), '&amp;'),
                      (("<"), '&lt;' ),
                      ((">"), '&gt;' ),
                      (('"'), '&quot;'))): #"
        text=str(value)
        for re, name in character_entities:
            if find(text, re) >= 0: text=join(split(text, re), name)
        return text


def str_field(v):
    if type(v) is ListType:
        return map(str_field,v)

    if hasattr(v,'__class__') and v.__class__ is FieldStorage:
        v=v.value
    elif type(v) is not StringType:
        if hasattr(v,'file') and v.file: v=v.file
        elif hasattr(v,'value'): v=v.value
    return v


class FileUpload:
    '''\
    File upload objects

    File upload objects are used to represent file-uploaded data.

    File upload objects can be used just like files.

    In addition, they have a 'headers' attribute that is a dictionary
    containing the file-upload headers, and a 'filename' attribute
    containing the name of the uploaded file.
    '''

    def __init__(self, aFieldStorage):

        file=aFieldStorage.file
        if hasattr(file, '__methods__'): methods=file.__methods__
        else: methods= ['close', 'fileno', 'flush', 'isatty',
                        'read', 'readline', 'readlines', 'seek',
                        'tell', 'truncate', 'write', 'writelines']

        d=self.__dict__
        for m in methods:
            if hasattr(file,m): d[m]=getattr(file,m)

        self.headers=aFieldStorage.headers
        self.filename=aFieldStorage.filename
    

def parse_cookie(text,
                 result=None,
                 qparmre=regex.compile(
                     '\([\0- ]*'
                     '\([^\0- ;,=\"]+\)="\([^"]*\)\"'
                     '\([\0- ]*[;,]\)?[\0- ]*\)'
                     ),
                 parmre=regex.compile(
                     '\([\0- ]*'
                     '\([^\0- ;,=\"]+\)=\([^\0- ;,\"]*\)'
                     '\([\0- ]*[;,]\)?[\0- ]*\)'
                     ),
                 ):

    if result is None: result={}
    already_have=result.has_key

    try:
        if qparmre.match(text) >= 0:
            # Match quoted correct cookies
            name=qparmre.group(2)
            value=qparmre.group(3)
            l=len(qparmre.group(1))
        elif parmre.match(text) >= 0:
            # Match evil MSIE cookies ;)
            name=parmre.group(2)
            value=parmre.group(3)
            l=len(parmre.group(1))
        else:
            # this may be an invalid cookie.
            # We'll simply bail without raising an error
            # if the cookie is invalid.
            return result
            
    finally: pass

    if not already_have(name): result[name]=value

    return apply(parse_cookie,(text[l:],result))

import sys

# add class
class record:
    def __getattr__(self, key, default=None):
        if key in ('get', 'keys', 'items', 'values', 'copy', 'has_key'):
            return getattr(self.__dict__, key)
        raise AttributeError, key

    def __getitem__(self, key):
        return self.__dict__[key]
            
    def __str__(self):
        L1 = self.__dict__.items()
        L1.sort()
        return join(map(lambda item: "%s: %s" %item, L1), ", ") 

    __repr__ = __str__

# Flags
SEQUENCE=1
DEFAULT=2
RECORD=4
RECORDS=8
REC=RECORD|RECORDS
EMPTY=16
CONVERTED=32




"""quixote.session

Contains various classes related to session management:
SessionCollection, GlobalSession, and ApplicationState.
"""

__revision__ = "$Id: session.py,v 1.13 2000/10/20 17:45:34 gward Exp $"

import sys, string
from time import time, localtime, strftime, strptime, mktime
from whrandom import randint

from quixote import config

def debug(*args):
    for arg in args: print arg,
    print

# This is used both to generate cookies and parse them (when we need
# to figure out how old they are)
TIME_FORMAT = "%Y%m%d:%H%M%S"


class SessionCollection:
    """SessionCollection acts as a dictionary of all sessions, mapping
    session ID strings to individual session objects.  Session objects
    are instances of some class, usually the GlobalSession class.

    Most applications can just use this class directly; sessions will
    be kept in memory-based dictionaries, and will be lost when the
    Quixote process dies.  Alternatively an application can subclass
    SessionCollection to implement specific behaviour, such as persistence.
    """

    cookie_name = config.COOKIE_NAME
    cookie_path = config.COOKIE_PATH
    
    def __init__ (self):
        self.sessions = {}

    # Methods to be overridden by subclasses
    
    def abort(self):
        "Abort all changes to sessions"
        pass

    def commit(self):
        "Commit all changes to sessions"
        pass

    def new_session(self, request, response, id):
        "Return a new instance of the GlobalSession class."
        # Subclasses can override this to return an instance of a different
        # class.
        return GlobalSession(request, response, id)

    
    # Hmmm, looks like a dictionary... smells like a dictionary...
    def keys(self):
        return self.sessions.keys()

    def values(self):
        return self.sessions.keys()

    def items(self):
        return self.sessions.items()

    def get(self, key):
        return self.sessions.get(key)

    def __getitem__(self, key):
        return self.sessions[key]

    def __delitem__ (self, key):
        del self.sessions[key]


    # and here's one real dictionaries *should* have ;-)
    def sorted_keys (self):
        k = self.sessions.keys()

        # Only call sort() if it's present; when using a ZODB BTree,
        # the object returned by keys() doesn't have a sort() method
        if hasattr(k, 'sort'): k.sort()

        return k


    # Now the real methods for real use

    def get_session (self, request, response):
        """Fetch or create a GlobalSession object for the current
           session, and return it.  Uses the "MX_session" cookie to
           get the ID of an existing session; sets a value for this
           cookie if it's not found or doesn't correspond to an
           existing session."""
        
        # XXX when/how should we clean up old sessions?

        id = request.cookies.get(self.cookie_name)
        if id == "" or id == "*del*":
            id = None
        if id is not None:
            session = self.sessions.get(id)
            
        # Hmmm.  This logic provokes slightly odd cookie behaviour in the
        # browser.  Could be worth rethinking it.
        if id is None or session is None:

            # Generate a "session ID", which is just the value of the
            # session cookie we are about to drop on the user
            while id is None or self.sessions.has_key (id):
                # IP address + date/time + random number
                id = ("%s:%s:%08X%08X" %
                      (request['REMOTE_ADDR'],
                       strftime(TIME_FORMAT, localtime(time())),
                       randint(0, sys.maxint-1),
                       randint(0, sys.maxint-1)))
                    
            # If we get here, it means this is the first we have seen of
            # this user in this session (or they've been rejecting our
            # cookies, which means we'll keep coming back here).  So, we
            # create a "global session" object -- this is the session
            # information that crosses application boundaries -- and
            # then drop a cookie on them, so they can get at this global
            # session object (and, more importantly, the bits of
            # application state it contains) later.  We also store the
            # session cookie in the *request*, in case multiple requests
            # are made for it (or one of its application state objects)
            # before we finish this request.  (After this request, the
            # user should be carrying around the cookie anyways.)

            session = self.new_session (request, response, id)
            self.sessions[id] = session
            self.set_session_cookie(request, response, id)

        session.access_time = time()
        return session

    # get_session ()


    def set_session_cookie (self, request, response, session_id):
        assert self.sessions.has_key(session_id), \
               "invalid session ID %s" % `session_id`
        response.setCookie(self.cookie_name, session_id,
                           domain = config.COOKIE_DOMAIN,
                           path = self.cookie_path)
        request.cookies[self.cookie_name] = session_id

    def revoke_session_cookie (self, request, response, session_id):
        assert self.sessions.has_key(session_id), \
               "invalid session ID" % `session_id`
        assert request.cookies[self.cookie_name] == session_id, \
               "not in session %s" % `session_id`
        response.setCookie(self.cookie_name, "*del*",
                           domain = config.COOKIE_DOMAIN,
                           path = self.cookie_path,
                           max_age = 0)
        del request.cookies[self.cookie_name]


    def has_session (self, request, must_exist=0):
        """Return true if the current request already has a cookie
           identifying a global session.  If 'must_exist' is true, the
           cookie must correspond to a currently existing session;
           otherwise (the default), we just check for the existence of
           the session cookie and don't inspect its content at all."""

        id = request.cookies.get(self.cookie_name)
        if id is None:
            return 0
        if must_exist:
            return self.sessions.has_key (id)
        else:
            return 1

    # has_session ()


    def get_app_state (self, request, response, app_id):
        """Fetch or create an application state object for the
           application identified by 'app_id' and the current
           session."""

        session = self.get_session (request, response)
        return session.get_app_state (app_id)


    def del_sessions (self, ip=None, age=None):
        """Delete a subset of open sessions selected according to the
        keyword arguments.  Default values select (and delete) no
        sessions.  Sessions are selected by combining non-None keyword
        args with "and": eg. if 'ip' and 'age' are non-None then any
        session matching *both* the supplied IP address and age (hours
        since session creation) is deleted.
        
        'ip' specifies an IP address (dotted-quad string); it must match
        the IP address of a session exactly.  'age' is the session's age
        (since creation) in hours; any session older than 'age' will
        pass the age test. Again, sessions must pass all specified tests
        (non-None arguments) to be selected and deleted."""
        
        goners = []
        for id in self.sessions.keys():
            (s_ip, s_date, s_time, rand) = string.split (id, ':')
            if ip is not None and ip != s_ip:
                continue
            if age is not None:
                datetime = strptime ("%s:%s" % (s_date, s_time), TIME_FORMAT)
                datetime = mktime (datetime)
            if (time() - datetime) <= float (age):
                continue

            # passed all the tests: we'll delete this sucker
            goners.append (id)

        return goners

    # del_session

# SessionCollection


class GlobalSession:

    """Holds session information that crosses application boundaries,
    and maintains a registry of application state objects where
    individual applications keep their particular session-state
    information.

    Instance attributes:
      id : string
        the session ID (generated by SessionCollection and used as the
        value of the session cookie)
      actual_user : any
        an object to identify the human being on the other end of the
        line.  It's up to you whether to store just a string in 'user',
        or some more complex data structure or object.
      user : any
        the user object used for access control checks.  Usually .user ==
        .actual_user, but for a situation where a user can act on
        behalf of another, you would leave .actual_user alone and
        set .user to another user object.
      creation_time : int
      access_time : int
        two ways of keeping track of the "age" of the whole session
        (without knowing anything about the age of individual
        application state objects).  Note that 'access_time' is
        maintained by the SessionCollection that owns this session.
      app_state : {string : ApplicationState}
        maps application IDs to application state objects (which should
        be instances of ApplicationState subclasses)
      app_creation_time : {string : int}
      app_access_time : {string : int}
        keep track of the creation and last-access time for each
        application state object in 'app_state'

    Feel free to access 'id', 'user', 'creation_time', and 'access_time'
    directly.  'id' and the '_time' attributes must never change;
    if 'user' or 'actual_user' need to change, use 'set_user()' or
    'set_actual_user()'.  All other attributes are private: do not
    access or modify them. 
    """


    # 'app_state_class' - registry of application state classes (so
    # we know how to create the objects that go in the per-instance
    # application state object registry)
    app_state_class = {}


    def __init__ (self, request, response, id):

        self.id = id
        self.user = self.actual_user = None
        self.creation_time = self.access_time = time()
        
        self.app_state = {}
        self.app_creation_time = {}
        self.app_access_time = {}

        debug ("GlobalSession: just created session %s, ctime %s" %
               (self.id,
                strftime ("%Y-%m-%d %H:%M:%S ",
                          localtime (self.creation_time))))
        #debug ("  URL0 = %s" % request.URL0)
        #debug ("  REFERER = %s" % request.HTTP_REFERER)

    # __init__ ()


    def __repr__ (self):
        return "<%s at %x: %s>" % (self.__class__.__name__, id(self), self.id)

    def __str__ (self):
        if self.actual_user is None:
            return "session %s (no user)" % self.id
        elif self.user == self.actual_user:
            return "session %s (user %s)" % (self.id, self.user)
        else:
            return "session %s (user %s working as %s)" % (self.id,
                                                         self.actual_user,
                                                         self.user)
    

    def dump (self, indent='', deep=1, header=1):
        time_fmt = "%Y-%m-%d %H:%M:%S"
        ctime = strftime (time_fmt, localtime (self.creation_time))
        atime = strftime (time_fmt, localtime (self.access_time))

        if header:
            lines = [indent + 'session %s:' % self.id]
            indent = indent + '  '
        else:
            lines = []

        lines.extend (
            [indent + 'user %s' % (self.user),
             indent + 'created %s, last accessed %s' % (ctime, atime)])

        if deep:
            apps = self.app_state.keys()
            if apps:
                lines.append (indent + 'application state objects:')
                apps.sort()
                indent = indent + '  '
                for app in apps:
                    ctime = self.app_creation_time.get(app)
                    if ctime is not None:
                        ctime = strftime (time_fmt, localtime (ctime))
                    atime = self.app_access_time.get(app)
                    if atime is not None:
                        atime = strftime (time_fmt, localtime (atime))

                    lines.append (indent +
                                  '%s: created %s, last accessed %s' %
                                  (app, ctime, atime))
                    app_state = self.app_state[app]
                    try:
                        st_lines = app_state.dump (indent + '  ', header=0)
                    except (AttributeError, NotImplementedError):
                        lines[-1] = lines[-1] + ": no 'dump' method"
                    else:
                        lines.extend (st_lines)

                # for app
            else:
                lines.append (indent + 'no application state objects')
        # if deep

        return lines
    
    # dump()
                            
            
    # -- Simple accessors and modifiers --------------------------------

    def set_actual_user (self, user):
        self.user = self.actual_user = user

    def set_user (self, user):
        self.user = user

    def get_creation_age (self):
        """Return the number of seconds since session was created."""
        return time() - self.creation_time

    def get_access_age (self):
        """Return the number of seconds since session was last accessed."""
        return time() - self.access_time


    # -- More involved stuff -------------------------------------------

    def get_app_state (self, app_id):

        state = self.app_state.get(app_id)
        if state is None:
            klass = self.app_state_class.get(app_id)
            if klass is None:
                raise RuntimeError, \
                      "no application state class has been registered " + \
                      "for application '%s'" % app_id

            state = klass ()
            # XXX restore this check (using mems.lib.base.isinstance?)!
            if 0 and not isinstance (state, ApplicationState):
                raise RuntimeError, \
                      ("class %s (for application '%s') " +
                       "not a subclass of ApplicationState") % (klass, app_id)

            self.app_state[app_id] = state
            self.app_creation_time[app_id] = \
                self.app_access_time[app_id] = time()

        else:
            self.app_access_time[app_id] = time()

        return state

    # get_app_state ()

# GlobalSession


class ApplicationState:
    """Defines the interface that application state classes must
       implement."""

    
    def __init__ (self):
        """Constructor takes no arguments, and just initializes all
           application session state structures."""

        raise NotImplementedError, \
              "ApplicationState should not be instantiated directly"

    def cleanup (self):
        """Destroy all application state data in memory without regard
           to persistence: the goal here is to free up memory, not to
           commit session data anywhere."""

        raise NotImplementedError, \
              "ApplicationState subclasses must implement 'cleanup()' method."

    def dump (self, indent='', header=1, deep=1):
        """Dump all relevant and interesting information in this application
           state object.  If 'header' is true (the default), the first line
           of the dump will identify the object, and the rest of the
           dump will be indented deeper than the header line.  Otherwise,
           the initial line is skipped and the whole dump is indented by
           'indent'.  If 'deep' is true, we'll dump contained objects
           that might be bulky or non-essential; otherwise, we'll skip
           dumping them."""

        raise NotImplementedError, \
              "ApplicationState subclasses must implement 'dump()' method."
        

# ApplicationState


def register_app_state_class (klass):
    try:
        app_id = klass.app_id
    except AttributeError:
        raise RuntimeError, \
              "application state class '%s' must define 'app_id' attribute" % \
              klass.__name__
    GlobalSession.app_state_class[app_id] = klass


#
# Pair of functions for getting and setting the singleton SessionCollection
# object.
# 

_session_coll = None

def get_session_collection ():
    global _session_coll
    if _session_coll is None:
        raise RuntimeError, "No session collection object has been set"
    
    return _session_coll

def set_session_collection (collection):
    global _session_coll
    if _session_coll is not None:
        raise RuntimeError, "Session collection object has already been set"
    _session_coll = collection
    
# These are the functions commonly used by callers.
def get_session (request, response):
    return get_session_collection().get_session (request, response)

def has_session (request=None, must_exist=0):
    return get_session_collection().has_session (request, must_exist)

def get_app_state (request, response, app_id):
    if app_id is None:
        raise RuntimeError, "app_id must be supplied"
    return get_session_collection().get_app_state (request, response, app_id)



"""quixote.config

Configuration constants
"""

__revision__ = "$Id: config.py,v 1.13 2000/10/02 21:32:04 gward Exp $"

# Name of root package for the Web site

ROOT_PACKAGE = 'mems.ui'

# Prefix of URLs that will be directed to Quixote
URL_PREFIX = '/q'

# E-mail address to send errors to
# (None to send no mail at all)

ERROR_EMAIL = None
#ERROR_EMAIL = 'akuchlin@mems-exchange.org'

# Filename for logging debugging output (None for no log file.)

#DEBUG_LOG = None
DEBUG_LOG = "/tmp/quixote-debug.log"

# Page shown when there's an access-control violation
ACCESS_VIOLATION_PAGE = '/access_error'

# Filename for logging error messages (None for output to standard error,
# which should wind up in the Web servers log file.)

#ERROR_LOG = None
ERROR_LOG = "/www/httpd/logs/quixote.log"

# If true, exceptions will also be returned to the browser; otherwise,
# a generic error display will be used.

DISPLAY_EXCEPTIONS = 1

# If true, Quixote will service exactly one request at a time, and
# then exit.  This makes no difference when you're running as a
# straight CGI script, but it makes it easier to debug while running
# as a FastCGI script.
RUN_ONCE = 0

# If true, then any "resource not found" errors will result in a
# consistent, terse, mostly-useless message.  If false, then the
# exact cause of failure will be returned.
SECURE_ERRORS = 1

# Session-related variables
# =========================

# Domain to specify in setting cookies to hold the session ID
COOKIE_DOMAIN = '.cnri.reston.va.us'

# Name and path of the cookie that will hold the ID
COOKIE_NAME = "MX_session"
COOKIE_PATH = "/"



"""quixote.publish

Logic for publishing modules and objects on the Web.
"""

__revision__ = "$Id: publish.py,v 1.42 2000/10/03 21:48:38 nascheme Exp $"

import sys, traceback, os, string, cStringIO
import time, types

from quixote import init, session, errors
from quixote.util import fcgi
from quixote.zope import HTTPRequest, HTTPResponse
from quixote import config


def get_module(name):
    """Get a module object by name."""
    __import__(name)
    module = sys.modules[name]
    return module
    
def _get_module_filename(module):
    filename = module.__file__
    if os.path.isdir( filename ):
        filename = os.path.join( filename, '__init__.py')
    elif filename[ -4: ] == '.pyc': filename = filename[:-1]

    return filename

def _traverse (original_path, path, request, response):
    """_traverse(original_path : string,
                 path : [string],
                 request, response : various objects) ->
          (script_name: string, path_info : string, object : any)

    Perform traversal based on the provided path, starting at the root
    object.  It returns the script name and path info values for
    the arrived-at object, along with the object itself.

    It's expected that the final object is something callable like a
    function or a method; intermediate objects along the way will
    usually be packages or modules.

    To prevent crackers from writing URLs that traverse private
    objects, every package, module, or object along the way must have
    an __exports__ attribute containing a list of publicly visible
    names.  Not having an __exports__ attribute is an error, though
    having __exports__ be an empty list is OK (though useless).  If a
    component of the path isn't in __exports__, that also produces an
    error.
    """

    # Traverse starting at the root
    object = get_module(config.ROOT_PACKAGE)

    # Loop over the components of the path, checking at every step
    # 1) whether __exports__ is present, 2) whether the path component
    # is listed in __exports__

    index = 0
    while index < len(path):
        component = path[index]

        # First security check: if the current doesn't even have an
        # __exports__ list, fail now: all Quixote-traversable namespaces
        # (modules, packages, instances) must have an export list!
        if not hasattr(object, '__exports__'):
            raise errors.TraversalError(
                "%s has no __exports__ list" % repr(object),
                original_path)

        # Second security check: make sure the current name component is
        # in the export list.  If not, there's a backdoor around the
        # usual traversal algorithm: namespaces may provide a
        # '__getname__()' function/method, which translates an arbitrary
        # string into an object that we continue traversing.  This is
        # very handy; it lets you put user-space objects into your
        # URL-space, eliminating the need for digging ID strings out of
        # a query, or checking PATHINFO after Quixote's done with it.
        # But it is a compromise to security: it opens up the traversal
        # algorithm to arbitrary names not listed in __exports__!
        if component not in object.__exports__:
            getname = getattr(object, '__getname__', None)
            if getname is not None:
                try:
                    object = getname(component)
                    if callable(object):
                        break
                except ValueError, msg:
                    raise errors.TraversalError(msg, original_path)
                index = index + 1
                continue
            else:
                raise errors.TraversalError(
                    ("%s not in __exports__ list for %s" %
                     (repr(component), object)),
                    original_path)

        # Call __access__ function or method, if it's present
        if hasattr(object, '__access__'):
            access_func = getattr(object, '__access__')
            # will raise AccessError if access failed
            access_func(request, response)
            
        is_module = type(object) is types.ModuleType
        has_component = hasattr(object, component)
        if is_module and not has_component:
            # get next module
            mod_name = object.__name__ + '.' + component
            try:
                object = get_module(mod_name)
            except ImportError, msg:
                raise errors.TraversalError(
                    (("module %s has no attribute '%s', "
                      "and failed to import %s: %s") %
                     (object.__name__, component, mod_name, msg)),
                    original_path)

        elif has_component:             # and 'object' is not a module
            object = getattr(object, component)

        else:                           # not a module, doesn't  'component'
            raise errors.TraversalError(
                "object %s has no attribute '%s'" % (`object`, component),
                original_path)

        # Bail out of the 'while' loop early once we reach something
        # callable (eg. a method or function).
        if callable(object):
            break

        index = index + 1

    # Determine the script name and path info to be returned.
    # If path_info is not the empty string, it must begin with a
    # leading slash, to guarantee that
    # "http://$SERVER_NAME:$SERVER_PORT$SCRIPT_NAME$PATH_INFO" will
    # always be an accessible URL that points to the current script.
    # This invariant comes from the cgi_path.html file in the Apache docs.
    script_name = config.URL_PREFIX + '/' + string.join(path[:index+1], '/')
    path_info =   string.join(path[index+1:], '/')
    if path_info != "": path_info = '/' + path_info 
    
    return script_name, path_info, object


def traverse_url (request, response):
    """traverse_url() -> (script_name, path_info, object)

    Reads the PATH_INFO environment variable and traverses the path
    found there.

    If the object ends up being a module, it's checked for a function
    named 'index', and that function will be returned instead.
    """

    # Get path and split it apart; /foo/bar/baz -> ['foo', 'bar', 'baz']
    original_path = os.environ.get('PATH_INFO', '')
    path = string.split(original_path, '/')
    path = filter(None, path)              # Drop empty strings
    
    (script_name, path_info, object) = \
        _traverse(original_path, path, request, response)

    # ExtensionClass-proof test for "this is a class instance"
    is_instance = (hasattr(object, '__class__') and
                   not hasattr(object, '__bases__'))
    is_module = type(object) == types.ModuleType

    if is_module:
        # It's a module, which isn't callable.  First, make sure URL has
        # trailing slash
        if original_path[-1] != '/':
            # repair path and redirect
            path = "%s/%s/" % (config.URL_PREFIX, string.join(path, '/'))
            def redirect_slash(request, response, path=path):
                return response.redirect(path)
            return '', '', redirect_slash
        
        # Check if the module has a function named 'index', and use that
        # function instead
        if (hasattr(object, 'index') and
            callable(object.index) and
            'index' in object.__exports__):
            object = object.index
        else:
            raise errors.TraversalError(
                "module %s has no function index (or it's not in __exports__)"
                % object.__name__,
                original_path)

    # Instances get a similar treatment to modules, except they may be
    # callable already (in which case we leave them be, and just call
    # them).
    elif is_instance:
        if callable(object):
            pass                        # object is just fine as-is
        elif (hasattr(object, 'index') and
              callable(object.index) and
              'index' in object.__exports__):
            object = object.index
        else:
            raise errors.TraversalError(
                ("object %s is not callable and has no 'index' method "
                 " (or 'index' not in __exports__)") % `object`,
                original_path)

    return script_name, path_info, object

def send_traceback (traceback_msg, request, response):
    """send_traceback(traceback_msg : string,
                      request : HTTPRequest,
                      response : HTTPResponse)

    If Quixote has been configured to do so, send an e-mail notifying
    someone of a traceback. 
    """
    
    if config.ERROR_EMAIL is None: return
    
    pipe = os.popen('/usr/lib/sendmail -f "Quixote" %s' % config.ERROR_EMAIL, 'w')
    pipe.write('Subject: Traceback\n')
    pipe.write('\n')

    pipe.write('Traceback\n=========\n')
    traceback.print_exc( file = pipe)

    if request is None:
        pipe.write('\nRequest is None\n')
    else:
        pipe.write('\nRequest\n=======\n')
        pipe.write( request.dump() )

    if response is None:
        pipe.write('\nResponse is None\n')
    else:
        pipe.write('\nResponse\n=======\n')
        pipe.write( str(response) )

    pipe.close()

exit_now = 0
    
def call (stdin, stdout, stderr, *args):
    """call(stdin : file, stdout : file, stderr : file, ...)

    The master function that does all the work for a single request.
    call() uses traverse_url() to get a callable object, creates the
    'request' and 'response' objects, and calls the object.  If it
    raises an exception, the traceback is logged and e-mailed.

    Any extra arguments will be passed to the callable.
    """
    global exit_now

    output = request = response = None
    sess_coll = session.get_session_collection()
    
    # Build request, response objects
    response = HTTPResponse.HTTPResponse( server_name = os.environ['SERVER_NAME'] )
    request = HTTPRequest.HTTPRequest( stdin, os.environ, response )
    request.processInputs()

    try:
        # Get the session object and stick it onto the request
        request.session = session.get_session (request, response)

        # Traverse package to a (hopefully-) callable object
        (os.environ['SCRIPT_NAME'],
         os.environ['PATH_INFO'],
         object) = traverse_url(request, response)

        if not callable(object):
            raise RuntimeError, "object not callable: %s" % repr(object)

        # Call object
        try:
            args = (request, response) + args
            output = apply(object, args)
        except SystemExit:
            output = "Server restarted"
            exit_now = 1

        if output is None:
            raise RuntimeError, 'Callable %s returned None' % repr(object)

    except errors.IgnoredException:
        # The callable ran OK, so return its output to the client
        sess_coll.commit()               # Commit any changes to the session

        location = response.headers['location']
        output = '<a href="%s">Follow this link</a>' % location
        response.setBody ( output )
        
        # Output results from Response object
        stdout.write( str( response ) )
        
    except errors.AccessError, msg:
        sess_coll.abort()
        if os.environ.get('PATH_INFO') == config.ACCESS_VIOLATION_PAGE:
            # recursion, eak!
            raise RuntimeError, "access violation page raised AccessError"
        os.environ['PATH_INFO'] = config.ACCESS_VIOLATION_PAGE
        call(stdin, stdout, stderr, msg) 

    except:
        traceback_file = cStringIO.StringIO()
        traceback.print_exc( file = traceback_file)

        if config.DISPLAY_EXCEPTIONS:
            stdout.write('Content-type: text/plain\n\n')
            stdout.write( traceback_file.getvalue() )

        tb = traceback_file.getvalue()
        if config.ERROR_LOG is None:
            sys.stderr.write( tb )
        else:
            try:
                errlog = open(config.ERROR_LOG, 'a')
            except IOError, exc:
                sys.stderr.write("error opening error log %s: %s\n" 
                                 % (`config.ERROR_LOG`, exc.strerror))
            else:
                errlog.write(tb)
                errlog.close()
            
        send_traceback( tb, request, response )

        # Abort any changes to the session
        sess_coll.abort()

    else:
        # The callable ran OK, so return its output to the client
        sess_coll.commit()               # Commit any changes to the session
        response.setBody ( output )

        # Output results from Response object
        stdout.write( str( response ) )


def call_fcgi ():
    """call_fcgi() -> None

    Entry point from CGI scripts; it will repeatedly do the call()
    function until there are no more FCGI requests.  This function can
    also be called in regular CGI scripts, in which case call() is
    called exactly once.
    """
    while fcgi.isFCGI() and not exit_now:
        f = fcgi.FCGI()
        os.environ = f.env
        call(f.inp, f.out, f.err)
        f.Finish()
        if config.RUN_ONCE: break
